from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RefType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    RFT_GENERAL: _ClassVar[RefType]
    RFT_USER: _ClassVar[RefType]
    RFT_DEVICE: _ClassVar[RefType]
    RFT_REC: _ClassVar[RefType]
    RFT_FOLDER: _ClassVar[RefType]
    RFT_TEAM: _ClassVar[RefType]
    RFT_ENTERPRISE: _ClassVar[RefType]
    RFT_PAM_DIRECTORY: _ClassVar[RefType]
    RFT_PAM_MACHINE: _ClassVar[RefType]
    RFT_PAM_DATABASE: _ClassVar[RefType]
    RFT_PAM_USER: _ClassVar[RefType]
    RFT_PAM_NETWORK: _ClassVar[RefType]
    RFT_PAM_BROWSER: _ClassVar[RefType]
    RFT_CONNECTION: _ClassVar[RefType]
    RFT_WORKFLOW: _ClassVar[RefType]
    RFT_NOTIFICATION: _ClassVar[RefType]
    RFT_USER_INFO: _ClassVar[RefType]
    RFT_TEAM_INFO: _ClassVar[RefType]
    RFT_ROLE: _ClassVar[RefType]

class GraphSyncDataType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    GSE_DATA: _ClassVar[GraphSyncDataType]
    GSE_KEY: _ClassVar[GraphSyncDataType]
    GSE_LINK: _ClassVar[GraphSyncDataType]
    GSE_ACL: _ClassVar[GraphSyncDataType]
    GSE_DELETION: _ClassVar[GraphSyncDataType]

class GraphSyncActorType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    GSA_USER: _ClassVar[GraphSyncActorType]
    GSA_SERVICE: _ClassVar[GraphSyncActorType]
    GSA_PAM_GATEWAY: _ClassVar[GraphSyncActorType]
RFT_GENERAL: RefType
RFT_USER: RefType
RFT_DEVICE: RefType
RFT_REC: RefType
RFT_FOLDER: RefType
RFT_TEAM: RefType
RFT_ENTERPRISE: RefType
RFT_PAM_DIRECTORY: RefType
RFT_PAM_MACHINE: RefType
RFT_PAM_DATABASE: RefType
RFT_PAM_USER: RefType
RFT_PAM_NETWORK: RefType
RFT_PAM_BROWSER: RefType
RFT_CONNECTION: RefType
RFT_WORKFLOW: RefType
RFT_NOTIFICATION: RefType
RFT_USER_INFO: RefType
RFT_TEAM_INFO: RefType
RFT_ROLE: RefType
GSE_DATA: GraphSyncDataType
GSE_KEY: GraphSyncDataType
GSE_LINK: GraphSyncDataType
GSE_ACL: GraphSyncDataType
GSE_DELETION: GraphSyncDataType
GSA_USER: GraphSyncActorType
GSA_SERVICE: GraphSyncActorType
GSA_PAM_GATEWAY: GraphSyncActorType

class GraphSyncRef(_message.Message):
    __slots__ = ["type", "value", "name"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    type: RefType
    value: bytes
    name: str
    def __init__(self, type: _Optional[_Union[RefType, str]] = ..., value: _Optional[bytes] = ..., name: _Optional[str] = ...) -> None: ...

class GraphSyncActor(_message.Message):
    __slots__ = ["type", "id", "name", "effectiveUserId"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    EFFECTIVEUSERID_FIELD_NUMBER: _ClassVar[int]
    type: GraphSyncActorType
    id: bytes
    name: str
    effectiveUserId: bytes
    def __init__(self, type: _Optional[_Union[GraphSyncActorType, str]] = ..., id: _Optional[bytes] = ..., name: _Optional[str] = ..., effectiveUserId: _Optional[bytes] = ...) -> None: ...

class GraphSyncData(_message.Message):
    __slots__ = ["type", "ref", "parentRef", "content", "path"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    REF_FIELD_NUMBER: _ClassVar[int]
    PARENTREF_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    type: GraphSyncDataType
    ref: GraphSyncRef
    parentRef: GraphSyncRef
    content: bytes
    path: str
    def __init__(self, type: _Optional[_Union[GraphSyncDataType, str]] = ..., ref: _Optional[_Union[GraphSyncRef, _Mapping]] = ..., parentRef: _Optional[_Union[GraphSyncRef, _Mapping]] = ..., content: _Optional[bytes] = ..., path: _Optional[str] = ...) -> None: ...

class GraphSyncDataPlus(_message.Message):
    __slots__ = ["data", "timestamp", "actor"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    ACTOR_FIELD_NUMBER: _ClassVar[int]
    data: GraphSyncData
    timestamp: int
    actor: GraphSyncActor
    def __init__(self, data: _Optional[_Union[GraphSyncData, _Mapping]] = ..., timestamp: _Optional[int] = ..., actor: _Optional[_Union[GraphSyncActor, _Mapping]] = ...) -> None: ...

class GraphSyncQuery(_message.Message):
    __slots__ = ["streamId", "origin", "syncPoint", "maxCount"]
    STREAMID_FIELD_NUMBER: _ClassVar[int]
    ORIGIN_FIELD_NUMBER: _ClassVar[int]
    SYNCPOINT_FIELD_NUMBER: _ClassVar[int]
    MAXCOUNT_FIELD_NUMBER: _ClassVar[int]
    streamId: bytes
    origin: bytes
    syncPoint: int
    maxCount: int
    def __init__(self, streamId: _Optional[bytes] = ..., origin: _Optional[bytes] = ..., syncPoint: _Optional[int] = ..., maxCount: _Optional[int] = ...) -> None: ...

class GraphSyncResult(_message.Message):
    __slots__ = ["streamId", "syncPoint", "data", "hasMore"]
    STREAMID_FIELD_NUMBER: _ClassVar[int]
    SYNCPOINT_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    HASMORE_FIELD_NUMBER: _ClassVar[int]
    streamId: bytes
    syncPoint: int
    data: _containers.RepeatedCompositeFieldContainer[GraphSyncDataPlus]
    hasMore: bool
    def __init__(self, streamId: _Optional[bytes] = ..., syncPoint: _Optional[int] = ..., data: _Optional[_Iterable[_Union[GraphSyncDataPlus, _Mapping]]] = ..., hasMore: bool = ...) -> None: ...

class GraphSyncMultiQuery(_message.Message):
    __slots__ = ["queries"]
    QUERIES_FIELD_NUMBER: _ClassVar[int]
    queries: _containers.RepeatedCompositeFieldContainer[GraphSyncQuery]
    def __init__(self, queries: _Optional[_Iterable[_Union[GraphSyncQuery, _Mapping]]] = ...) -> None: ...

class GraphSyncMultiResult(_message.Message):
    __slots__ = ["results"]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    results: _containers.RepeatedCompositeFieldContainer[GraphSyncResult]
    def __init__(self, results: _Optional[_Iterable[_Union[GraphSyncResult, _Mapping]]] = ...) -> None: ...

class GraphSyncAddDataRequest(_message.Message):
    __slots__ = ["origin", "data"]
    ORIGIN_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    origin: GraphSyncRef
    data: _containers.RepeatedCompositeFieldContainer[GraphSyncData]
    def __init__(self, origin: _Optional[_Union[GraphSyncRef, _Mapping]] = ..., data: _Optional[_Iterable[_Union[GraphSyncData, _Mapping]]] = ...) -> None: ...

class GraphSyncLeafsQuery(_message.Message):
    __slots__ = ["vertices"]
    VERTICES_FIELD_NUMBER: _ClassVar[int]
    vertices: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, vertices: _Optional[_Iterable[bytes]] = ...) -> None: ...

class GraphSyncRefsResult(_message.Message):
    __slots__ = ["refs"]
    REFS_FIELD_NUMBER: _ClassVar[int]
    refs: _containers.RepeatedCompositeFieldContainer[GraphSyncRef]
    def __init__(self, refs: _Optional[_Iterable[_Union[GraphSyncRef, _Mapping]]] = ...) -> None: ...



from collections import OrderedDict
import sys
from unittest import TestCase, mock

from keepercommander.commands import base
from keepercommander.cli import do_command, read_command_with_continuation
from data_vault import get_connected_params


class TestCommandLineInterface(TestCase):

    def test_command_import(self):
        commands = {}
        aliases = {}
        command_info = OrderedDict()

        #base.register_commands(commands, aliases, command_info)
        base.register_enterprise_commands(commands, aliases, command_info)

    def test_normalize_output_param(self):
        saved_platform = sys.platform
        try:
            # simulate windows platform
            sys.platform = 'win32'
            s = base.normalize_output_param(r'command --output=d:\1\2\aaa')
            self.assertEqual(s, r'command --output=d:/1/2/aaa')

            s = base.normalize_output_param(r'command --output d:\1\2\aaa')
            self.assertEqual(s, r'command --output d:/1/2/aaa')

            s = base.normalize_output_param(r'command d:\1\2\aaa')
            self.assertEqual(s, r'command d:\1\2\aaa')

            s = base.normalize_output_param(r'command --output d:/1/2\ 3/aaa')
            self.assertEqual(s, r'command --output d:/1/2\ 3/aaa')

            # simulate osx platform
            sys.platform = 'darwin'
            s = base.normalize_output_param(r'command --output=d:\1\2\aaa')
            self.assertEqual(s, r'command --output=d:\1\2\aaa')            
        finally:
            sys.platform = saved_platform

    def test_do_command_no_opts(self):
        params = get_connected_params()
        params.sync_data = False
        with mock.patch('keepercommander.commands.utils.ThisDeviceCommand.print_device_info') \
                as mock_print_dev:
            mock_print_dev.return_value = 'test device info'
            do_command(params, 'this-device')
            mock_print_dev.assert_called()

    def test_line_continuation(self):
        """Test that line continuation with backslash works correctly."""
        params = get_connected_params()
        
        # Mock input to simulate line continuation with varying whitespace
        input_lines = [
            'record-add -t "Test Record" \\',
            '  -rt login \\',  # Leading whitespace 
            '  login=testuser \\',  # Leading whitespace
            '  password=testpass'  # Final line with leading whitespace
        ]
        
        with mock.patch('builtins.input', side_effect=input_lines):
            result = read_command_with_continuation(None, params)
            expected = 'record-add -t "Test Record" -rt login login=testuser password=testpass'
            self.assertEqual(result, expected)

    def test_line_continuation_no_backslash(self):
        """Test that commands without line continuation work normally."""
        params = get_connected_params()
        
        with mock.patch('builtins.input', return_value='record-add -t "Simple Test" -rt login'):
            result = read_command_with_continuation(None, params)
            expected = 'record-add -t "Simple Test" -rt login'
            self.assertEqual(result, expected)

    def test_line_continuation_with_empty_lines(self):
        """Test that empty continuation lines are handled correctly."""
        params = get_connected_params()
        
        # Mock input with empty continuation lines
        input_lines = [
            'record-add -t "Test Record" \\',
            '  \\',  # Empty line with just backslash - should be skipped
            '  -rt login \\',
            'login=testuser'  # Final line without backslash
        ]
        
        with mock.patch('builtins.input', side_effect=input_lines):
            result = read_command_with_continuation(None, params)
            expected = 'record-add -t "Test Record" -rt login login=testuser'
            self.assertEqual(result, expected)

    def test_line_continuation_with_trailing_spaces(self):
        """Test that line continuation handles trailing spaces after backslash gracefully."""
        params = get_connected_params()
        
        # Mock input with trailing spaces after backslashes (common user error)
        input_lines = [
            'record-add -t "Gmail Account" -rt login \\ ',   # space after backslash
            '  login=john.doe@gmail.com \\  ',              # spaces after backslash
            '  password=SecurePass123 \\	',               # tab after backslash
            '  url=https://accounts.google.com'
        ]
        
        with mock.patch('builtins.input', side_effect=input_lines):
            result = read_command_with_continuation(None, params)
            expected = 'record-add -t "Gmail Account" -rt login login=john.doe@gmail.com password=SecurePass123 url=https://accounts.google.com'
            self.assertEqual(result, expected)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from flask import Flask
from typing import Optional
from .command import create_command_blueprint, create_legacy_command_blueprint
from ..decorators.logging import logger, debug_decorator

def _setup_queue_mode(app: Flask) -> None:
    """Setup queue mode with v2 API endpoints."""
    from ..core.request_queue import queue_manager
    queue_manager.start()
    
    command_bp = create_command_blueprint()
    app.register_blueprint(command_bp, url_prefix='/api/v2')
    logger.debug("Started queue manager and registered command blueprint with URL prefix '/api/v2'")

def _setup_legacy_mode(app: Flask) -> None:
    """Setup legacy mode with v1 API endpoints."""
    legacy_bp = create_legacy_command_blueprint()
    app.register_blueprint(legacy_bp, url_prefix='/api/v1')
    logger.info("Using /api/v1 - Enable queue mode (-q y) for /api/v2")

@debug_decorator
def init_routes(app: Optional[Flask] = None) -> None:
    """Initialize routes and queue manager for the Keeper Commander Service."""
    if app is None:
        raise ValueError("App instance is required")
    
    logger.debug("Starting route initialization")
    
    try:
        from ..config.service_config import ServiceConfig
        service_config = ServiceConfig()
        config_data = service_config.load_config()
        queue_enabled = config_data.get("queue_enabled", "y")  # Default to enabled
        
        if queue_enabled == "y":
            logger.debug("Queue enabled - setting up v2 API with request queue")
            _setup_queue_mode(app)
        else:
            logger.debug("Queue disabled - setting up v1 API with direct execution")
            _setup_legacy_mode(app)
            
    except Exception as e:
        logger.warning(f"Could not load service config, defaulting to queue mode: {e}")
        _setup_queue_mode(app)
    
    logger.debug("Route initialization completed successfully")



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
import logging
from typing import Any, List, Dict, Optional

import google

from . import api, utils, crypto, convert_keys
from .display import bcolors
from .params import KeeperParams, RecordOwner
from .proto import SyncDown_pb2, record_pb2, client_pb2, breachwatch_pb2
from .subfolder import RootFolderNode, UserFolderNode, SharedFolderNode, SharedFolderFolderNode, BaseFolderNode
from .vault import KeeperRecord


def sync_down(params, record_types=False):   # type: (KeeperParams, bool) -> None
    """Sync full or partial data down to the client"""

    params.sync_data = False
    token = params.sync_down_token
    if not token:
        logging.info('Syncing...')

    for record in params.record_cache.values():
        if 'shares' in record:
            del record['shares']

    def delete_record_key(rec_uid):
        if rec_uid in params.record_cache:
            record = params.record_cache[rec_uid]
            if 'record_key_unencrypted' in record:
                del record['record_key_unencrypted']
                if 'data_unencrypted' in record:
                    del record['data_unencrypted']
                if 'extra_unencrypted' in record:
                    del record['extra_unencrypted']

    def delete_shared_folder_key(sf_uid):
        if sf_uid in params.shared_folder_cache:
            shared_folder = params.shared_folder_cache[sf_uid]
            if 'shared_folder_key_unencrypted' in shared_folder:
                del shared_folder['shared_folder_key_unencrypted']
                if 'records' in shared_folder:
                    for sfr in shared_folder['records']:
                        record_uid = sfr['record_uid']
                        if record_uid not in params.meta_data_cache:
                            delete_record_key(record_uid)

    def delete_team_key(team_uid):
        if team_uid in params.team_cache:
            team = params.team_cache[team_uid]
            if 'team_key_unencrypted' in team:
                del team['team_key_unencrypted']
                if 'shared_folder_keys' in team:
                    for sfk in team['shared_folder_keys']:
                        delete_shared_folder_key(sfk['shared_folder_uid'])

    params.available_team_cache = None

    resp_bw_recs = []            # type: List[SyncDown_pb2.BreachWatchRecord]
    resp_sec_data_recs = []      # type: List[SyncDown_pb2.BreachWatchSecurityData]
    resp_sec_scores = []         # type: List[SyncDown_pb2.SecurityScoreData]
    request = SyncDown_pb2.SyncDownRequest()
    revision = params.revision
    full_sync = False
    done = False
    while not done:
        if token:
            request.continuationToken = token
        response = api.communicate_rest(params, request, 'vault/sync_down', rs_type=SyncDown_pb2.SyncDownResponse)
        done = not response.hasMore
        token = response.continuationToken
        if response.cacheStatus == SyncDown_pb2.CLEAR:
            full_sync = True
            params.record_cache.clear()
            params.record_rotation_cache.clear()
            params.meta_data_cache.clear()
            params.shared_folder_cache.clear()
            params.team_cache.clear()
            params.available_team_cache = None
            params.subfolder_cache.clear()
            params.subfolder_record_cache.clear()
            params.record_history.clear()
            params.record_owner_cache.clear()
            params.breach_watch_security_data.clear()
            params.breach_watch_records.clear()
            params.security_score_data.clear()

        if len(response.removedRecords) > 0:
            logging.debug('Processing removed records')
            for record_uid_bytes in response.removedRecords:
                record_uid = utils.base64_url_encode(record_uid_bytes)
                # remove BW record data
                if record_uid in params.breach_watch_records:
                    del params.breach_watch_records[record_uid]
                # remove associated security data
                if record_uid in params.breach_watch_security_data:
                    del params.breach_watch_security_data[record_uid]
                # remove associated security score data
                params.security_score_data.pop(record_uid, None)
                # remove record metadata
                if record_uid in params.meta_data_cache:
                    del params.meta_data_cache[record_uid]
                # delete record key
                delete_record_key(record_uid)
                # remove record from user folders
                for folder_uid in params.subfolder_record_cache:
                    if record_uid in params.subfolder_record_cache[folder_uid]:
                        if folder_uid in params.subfolder_cache:
                            folder = params.subfolder_cache[folder_uid]
                            if folder.get('type') == 'user_folder':
                                params.subfolder_record_cache[folder_uid].remove(record_uid)
                        elif folder_uid == '':
                            params.subfolder_record_cache[folder_uid].remove(record_uid)

        if len(response.removedTeams) > 0:
            logging.debug('Processing removed teams')
            for team_uid_bytes in response.removedTeams:
                team_uid = utils.base64_url_encode(team_uid_bytes)
                delete_team_key(team_uid)
                # remove team from shared folder
                for shared_folder_uid in params.shared_folder_cache:
                    shared_folder = params.shared_folder_cache[shared_folder_uid]
                    if 'teams' in shared_folder:
                        shared_folder['teams'] = [x for x in shared_folder['teams'] if x['team_uid'] != team_uid]
                if team_uid in params.team_cache:
                    del params.team_cache[team_uid]

        if len(response.removedSharedFolders) > 0:
            logging.debug('Processing removed shared folders')
            for sf_uid_bytes in response.removedSharedFolders:
                sf_uid = utils.base64_url_encode(sf_uid_bytes)
                if sf_uid in params.shared_folder_cache:
                    delete_shared_folder_key(sf_uid)
                    shared_folder = params.shared_folder_cache[sf_uid]
                    if 'shared_folder_key' in shared_folder:
                        del shared_folder['shared_folder_key']
                    if 'key_type' in shared_folder:
                        del shared_folder['key_type']
                    if 'users' in shared_folder:
                        shared_folder['users'] = [x for x in shared_folder['users'] if x['username'] != params.user]
                    if 'records' in shared_folder:
                        for r in shared_folder['records']:
                            if 'record_uid' in r:
                                delete_record_key(r['record_uid'])

        if len(response.removedRecordLinks) > 0:
            logging.debug('Processing removed record links')
            for link in response.removedRecordLinks:
                record_uid = utils.base64_url_encode(link.childRecordUid)
                if record_uid in params.record_cache:
                    delete_record_key(record_uid)
                    parents = params.record_link_cache[record_uid]
                    parent_uid = utils.base64_url_encode(link.parentRecordUid)
                    if parent_uid in parents:
                        del parents[parent_uid]

        if len(response.removedUserFolders) > 0:
            for f_uid_bytes in response.removedUserFolders:
                f_uid = utils.base64_url_encode(f_uid_bytes)
                if f_uid in params.subfolder_cache:
                    del params.subfolder_cache[f_uid]
                if f_uid in params.subfolder_record_cache:
                    del params.subfolder_record_cache[f_uid]

        if len(response.removedSharedFolderFolders) > 0:
            for sffr in response.removedSharedFolderFolders:
                f_uid_bytes = sffr.folderUid or sffr.sharedFolderUid
                f_uid = utils.base64_url_encode(f_uid_bytes)
                if f_uid in params.subfolder_cache:
                    del params.subfolder_cache[f_uid]
                if f_uid in params.subfolder_record_cache:
                    del params.subfolder_record_cache[f_uid]

        if len(response.removedUserFolderSharedFolders) > 0:
            for ufsfr in response.removedUserFolderSharedFolders:
                f_uid = utils.base64_url_encode(ufsfr.sharedFolderUid)
                if f_uid in params.subfolder_cache:
                    del params.subfolder_cache[f_uid]
                if f_uid in params.subfolder_record_cache:
                    del params.subfolder_record_cache[f_uid]

        if len(response.removedUserFolderRecords) > 0:
            for ufrr in response.removedUserFolderRecords:
                f_uid = utils.base64_url_encode(ufrr.folderUid) if ufrr.folderUid else ''
                if f_uid in params.subfolder_record_cache:
                    rs = params.subfolder_record_cache[f_uid]
                    r_uid = utils.base64_url_encode(ufrr.recordUid)
                    if r_uid in rs:
                        rs.remove(r_uid)

        if len(response.removedSharedFolderFolderRecords) > 0:
            for sfrr in response.removedSharedFolderFolderRecords:
                f_uid = utils.base64_url_encode(sfrr.folderUid or sfrr.sharedFolderUid)
                if f_uid in params.subfolder_record_cache:
                    rs = params.subfolder_record_cache[f_uid]
                    r_uid = utils.base64_url_encode(sfrr.recordUid)
                    if r_uid in rs:
                        rs.remove(r_uid)

        if len(response.recordLinks) > 0:
            for rl in response.recordLinks:
                child_uid = utils.base64_url_encode(rl.childRecordUid)
                parents = params.record_link_cache.get(child_uid)
                if parents is None:
                    parents = {}
                    params.record_link_cache[child_uid] = parents
                parent_uid = utils.base64_url_encode(rl.parentRecordUid)
                parents[parent_uid] = {
                    'child_uid': child_uid,
                    'parent_uid': parent_uid,
                    'record_key': utils.base64_url_encode(rl.recordKey)
                }

        if len(response.recordMetaData) > 0:
            logging.debug('Processing record_meta_data')
            for rmd in response.recordMetaData:
                meta_data = {
                    'record_uid': utils.base64_url_encode(rmd.recordUid),
                    'owner': rmd.owner,
                    'can_edit': rmd.canEdit,
                    'can_share': rmd.canShare,
                    'record_key': utils.base64_url_encode(rmd.recordKey),
                    'record_key_type': rmd.recordKeyType,
                    'owner_account_uid': utils.base64_url_encode(rmd.ownerAccountUid or params.account_uid_bytes),
                }  # type: dict
                if rmd.expiration > 0:
                    meta_data['expiration'] = rmd.expiration
                record_uid = meta_data['record_uid']
                params.meta_data_cache[record_uid] = meta_data
                params.record_owner_cache[record_uid] = RecordOwner(meta_data['owner'], meta_data['owner_account_uid'])

        if len(response.records) > 0:
            logging.debug('Processing records')
            r = max((x.revision for x in response.records))
            if r > revision:
                revision = r

            def convert_record(r):
                o = {
                    'record_uid': utils.base64_url_encode(r.recordUid),
                    'revision': r.revision,
                    'version': r.version,
                    'shared': r.shared,
                    'client_modified_time': r.clientModifiedTime,
                    'data': utils.base64_url_encode(r.data),
                    'extra': utils.base64_url_encode(r.extra),
                    'file_size': r.fileSize,
                    'thumbnail_size': r.thumbnailSize,
                }
                if r.udata:
                    o['udata'] = json.loads(r.udata)
                return o

            for r in response.records:
                record = convert_record(r)
                params.record_cache[record['record_uid']] = record

        if len(response.nonSharedData) > 0:
            for nsd in response.nonSharedData:
                record_uid = utils.base64_url_encode(nsd.recordUid)
                params.non_shared_data_cache[record_uid] = {
                    'record_uid': record_uid,
                    'data': utils.base64_url_encode(nsd.data),
                }

        if len(response.teams) > 0:
            def assign_team(t, team):
                team['name'] = t.name
                team['team_key'] = utils.base64_url_encode(t.teamKey)
                team['team_key_type'] = t.teamKeyType
                team['restrict_edit'] = t.restrictEdit
                team['restrict_share'] = t.restrictShare
                team['restrict_view'] = t.restrictView
                if len(t.teamEccPrivateKey) > 0:
                    team['team_ec_private_key'] = utils.base64_url_encode(t.teamEccPrivateKey)
                if len(t.teamPrivateKey) > 0:
                    team['team_private_key'] = utils.base64_url_encode(t.teamPrivateKey)

            for t in response.teams:
                team_uid = utils.base64_url_encode(t.teamUid)
                team = params.team_cache.get(team_uid)
                if team is None:
                    team = {'team_uid': team_uid}
                    params.team_cache[team_uid] = team
                assign_team(t, team)

                if len(t.removedSharedFolders) > 0 and 'shared_folder_keys' in team:
                    sf_keys = team['shared_folder_keys']
                    if isinstance(sf_keys, list):
                        for rsf in t.removedSharedFolders:
                            sf_uid = utils.base64_url_encode(rsf)
                            delete_shared_folder_key(sf_uid)
                            pos = next((i for i, x in enumerate(sf_keys) if x['shared_folder_uid'] == sf_uid), -1)
                            if pos >= 0:
                                del sf_keys[pos]

                if len(t.sharedFolderKeys) > 0:
                    if 'shared_folder_keys' not in team:
                        team['shared_folder_keys'] = []
                    sf_keys = team['shared_folder_keys']  # type: List[Dict]
                    for sfk in t.sharedFolderKeys:
                        sf_uid = utils.base64_url_encode(sfk.sharedFolderUid)
                        sf_key = next((x for x in sf_keys if x['shared_folder_uid'] == sf_uid), None)
                        if sf_key is None:
                            sf_key = {
                                'shared_folder_uid': sf_uid
                            }
                            sf_keys.append(sf_key)
                        sf_key['shared_folder_key'] = utils.base64_url_encode(sfk.sharedFolderKey)
                        sf_key['key_type'] = sfk.keyType

        if len(response.sharedFolders) > 0:
            logging.debug('Processing shared_folders')
            r = max((x.revision for x in response.sharedFolders))
            if r > revision:
                revision = r

            def assign_shared_folder(sf, o):
                o['revision'] = sf.revision
                o['name'] = utils.base64_url_encode(sf.name)
                o['data'] = utils.base64_url_encode(sf.data)
                o['default_manage_records'] = sf.defaultManageRecords
                o['default_manage_users'] = sf.defaultManageUsers
                o['default_can_edit'] = sf.defaultCanEdit
                o['default_can_share'] = sf.defaultCanReshare
                o['owner_account_uid'] = utils.base64_url_encode(sf.ownerAccountUid or params.account_uid_bytes)

            for p_sf in response.sharedFolders:
                shared_folder_uid = utils.base64_url_encode(p_sf.sharedFolderUid)
                shared_folder = params.shared_folder_cache.get(shared_folder_uid)
                if shared_folder is None:
                    shared_folder = {
                        'shared_folder_uid': shared_folder_uid
                    }
                    params.shared_folder_cache[shared_folder_uid] = shared_folder
                else:
                    if p_sf.cacheStatus == SyncDown_pb2.CLEAR:
                        if 'users' in shared_folder:
                            del shared_folder['users']
                        if 'teams' in shared_folder:
                            del shared_folder['teams']
                        if 'records' in shared_folder:
                            del shared_folder['records']
                    if 'shared_folder_key_unencrypted' in shared_folder:
                        del shared_folder['shared_folder_key_unencrypted']
                    if 'data_unencrypted' in shared_folder:
                        del shared_folder['data_unencrypted']
                assign_shared_folder(p_sf, shared_folder)

                if p_sf.sharedFolderKey:
                    shared_folder['shared_folder_key'] = utils.base64_url_encode(p_sf.sharedFolderKey)
                    shared_folder['key_type'] = p_sf.keyType
                else:
                    if 'shared_folder_key' in shared_folder:
                        del shared_folder['shared_folder_key']
                    if 'key_type' in shared_folder:
                        del shared_folder['key_type']

        if len(response.sharedFolderUsers) > 0:
            for sfu in response.sharedFolderUsers:
                shared_folder_uid = utils.base64_url_encode(sfu.sharedFolderUid)
                account_uid = utils.base64_url_encode(sfu.accountUid) if sfu.accountUid else utils.base64_url_encode(params.account_uid_bytes)
                if shared_folder_uid in params.shared_folder_cache:
                    sf = params.shared_folder_cache[shared_folder_uid]
                    if 'users' not in sf:
                        sf['users'] = []
                    sf_user = next((x for x in sf['users'] if x['account_uid'] == account_uid), None)
                    if sf_user is None:
                        sf_user = {
                            'username': sfu.username,
                            'account_uid': account_uid
                        }
                        sf['users'].append(sf_user)
                    sf_user['manage_records'] = sfu.manageRecords
                    sf_user['manage_users'] = sfu.manageUsers
                    if sfu.expiration > 0:
                        sf_user['expiration'] = sfu.expiration

        if len(response.sharedFolderTeams) > 0:
            for sft in response.sharedFolderTeams:
                shared_folder_uid = utils.base64_url_encode(sft.sharedFolderUid)
                if shared_folder_uid in params.shared_folder_cache:
                    sf = params.shared_folder_cache[shared_folder_uid]
                    if 'teams' not in sf:
                        sf['teams'] = []
                    team_uid = utils.base64_url_encode(sft.teamUid)
                    sf_team = next((x for x in sf['teams'] if x['team_uid'] == team_uid), None)
                    if sf_team is None:
                        sf_team = {
                            'team_uid': team_uid
                        }
                        sf['teams'].append(sf_team)
                    sf_team['name'] = sft.name if hasattr(sft, 'name') else ''
                    sf_team['manage_records'] = sft.manageRecords
                    sf_team['manage_users'] = sft.manageUsers
                    if sft.expiration > 0:
                        sf_team['expiration'] = sft.expiration

        if len(response.sharedFolderRecords) > 0:
            def assign_shared_folder_record(sfr, sf_record):
                sf_record['record_key'] = utils.base64_url_encode(sfr.recordKey)
                sf_record['can_share'] = sfr.canShare
                sf_record['can_edit'] = sfr.canEdit
                sf_record['owner'] = sfr.owner
                sf_record['owner_account_uid'] = utils.base64_url_encode(sfr.ownerAccountUid or params.account_uid_bytes)
                if sfr.expiration > 0:
                    sf_record['expiration'] = sfr.expiration

            for sfr in response.sharedFolderRecords:
                shared_folder_uid = utils.base64_url_encode(sfr.sharedFolderUid)
                if shared_folder_uid in params.shared_folder_cache:
                    sf = params.shared_folder_cache[shared_folder_uid]
                    if 'records' not in sf:
                        sf['records'] = []
                    record_uid = utils.base64_url_encode(sfr.recordUid)
                    sf_record = next((x for x in sf['records'] if x['record_uid'] == record_uid), None)  # type: Dict
                    if sf_record is None:
                        sf_record = {
                            'record_uid': record_uid
                        }
                        sf['records'].append(sf_record)
                    assign_shared_folder_record(sfr, sf_record)
                    params.record_owner_cache[record_uid] = \
                        RecordOwner(sf_record['owner'], sf_record['owner_account_uid'])

        if len(response.removedSharedFolderRecords) > 0:
            for rsfr in response.removedSharedFolderRecords:
                shared_folder_uid = utils.base64_url_encode(rsfr.sharedFolderUid)
                record_uid = utils.base64_url_encode(rsfr.recordUid)
                delete_record_key(record_uid)
                if shared_folder_uid in params.shared_folder_cache:
                    sf = params.shared_folder_cache[shared_folder_uid]  # type: dict
                    if 'records' in sf:
                        pos = next((i for i, x in enumerate(sf['records']) if x['record_uid'] == record_uid), -1)
                        if pos >= 0:
                            del sf['records'][pos]

        if len(response.removedSharedFolderUsers) > 0:
            for rsfu in response.removedSharedFolderUsers:
                shared_folder_uid = utils.base64_url_encode(rsfu.sharedFolderUid)
                if shared_folder_uid in params.shared_folder_cache:
                    sf = params.shared_folder_cache[shared_folder_uid]
                    if 'users' in sf:
                        if len(rsfu.username) > 0:
                            pos = next((i for i, x in enumerate(sf['users']) if x['username'] == rsfu.username), -1)
                        else:
                            account_uid = utils.base64_url_encode(rsfu.accountUid)
                            pos = next((i for i, x in enumerate(sf['users']) if x['account_uid'] == account_uid), -1)
                        if pos >= 0:
                            del sf['users'][pos]

        if len(response.removedSharedFolderTeams) > 0:
            for rsft in response.removedSharedFolderTeams:
                shared_folder_uid = utils.base64_url_encode(rsft.sharedFolderUid)
                team_uid = utils.base64_url_encode(rsft.teamUid)
                if shared_folder_uid in params.shared_folder_cache:
                    sf = params.shared_folder_cache[shared_folder_uid]
                    if 'teams' in sf:
                        pos = next((i for i, x in enumerate(sf['teams']) if x['team_uid'] == team_uid), -1)
                        if pos >= 0:
                            del sf['teams'][pos]

        if len(response.userFolders) > 0:
            def convert_user_folder(uf):
                o = {
                    'folder_uid': utils.base64_url_encode(uf.folderUid),
                    'user_folder_key': utils.base64_url_encode(uf.userFolderKey),
                    'key_type': uf.keyType,
                    'revision': uf.revision,
                    'type': 'user_folder',
                    'data': utils.base64_url_encode(uf.data),
                }
                if uf.parentUid:
                    o['parent_uid'] = utils.base64_url_encode(uf.parentUid)
                return o

            for uf in response.userFolders:
                user_folder = convert_user_folder(uf)
                params.subfolder_cache[user_folder['folder_uid']] = user_folder

        if len(response.userFolderRecords) > 0:
            for ufr in response.userFolderRecords:
                fuid = utils.base64_url_encode(ufr.folderUid) if ufr.folderUid else ''
                if fuid not in params.subfolder_record_cache:
                    params.subfolder_record_cache[fuid] = set()
                record_uid = utils.base64_url_encode(ufr.recordUid)
                params.subfolder_record_cache[fuid].add(record_uid)

        if len(response.userFolderSharedFolders) > 0:
            def convert_user_folder_shared_folder(ufsf):
                o = {
                    'shared_folder_uid': utils.base64_url_encode(ufsf.sharedFolderUid),
                    'revision': ufsf.revision,
                    'type': 'shared_folder'
                }
                if ufsf.folderUid:
                    is_fake_uid = ufsf.folderUid.endswith(b'\0\0\0\0\0\0\0\0\0\0') or \
                                  ufsf.folderUid.startswith(b'\0\0\0\0\0\0\0\0\0\0')
                    if not is_fake_uid:
                        o['folder_uid'] = utils.base64_url_encode(ufsf.folderUid)
                return o

            for ufsf in response.userFolderSharedFolders:
                uf_sf = convert_user_folder_shared_folder(ufsf)
                sf_uid = uf_sf['shared_folder_uid']
                params.subfolder_cache[sf_uid] = uf_sf

        if len(response.sharedFolderFolders) > 0:
            for p_sff in response.sharedFolderFolders:
                sff = {
                    'shared_folder_uid': utils.base64_url_encode(p_sff.sharedFolderUid),
                    'folder_uid': utils.base64_url_encode(p_sff.folderUid),
                    'revision': p_sff.revision,
                    'shared_folder_folder_key': utils.base64_url_encode(p_sff.sharedFolderFolderKey),
                    'key_type': p_sff.keyType,
                    'type': 'shared_folder_folder',
                    'data': utils.base64_url_encode(p_sff.data),
                }
                if p_sff.parentUid:
                    sff['parent_uid'] = utils.base64_url_encode(p_sff.parentUid)
                params.subfolder_cache[sff['folder_uid']] = sff

        if len(response.sharedFolderFolderRecords) > 0:
            for sffr in response.sharedFolderFolderRecords:
                key = utils.base64_url_encode(sffr.folderUid or sffr.sharedFolderUid)
                if key not in params.subfolder_record_cache:
                    params.subfolder_record_cache[key] = set()
                record_uid = utils.base64_url_encode(sffr.recordUid)
                params.subfolder_record_cache[key].add(record_uid)

        if len(response.sharingChanges) > 0:
            for sharing_change in response.sharingChanges:
                record_uid = utils.base64_url_encode(sharing_change.recordUid)
                if record_uid in params.record_cache:
                    record = params.record_cache[record_uid]
                    record['shared'] = sharing_change.shared

        if len(response.shareInvitations) > 0:
            params.pending_share_requests.update((x.username for x in response.shareInvitations))

        if len(response.breachWatchRecords) > 0:
            resp_bw_recs.extend(response.breachWatchRecords)

        if len(response.breachWatchSecurityData) > 0:
            resp_sec_data_recs.extend(response.breachWatchSecurityData)

        if len(response.securityScoreData) > 0:
            resp_sec_scores.extend(response.securityScoreData)

        if len(response.removedUsers) > 0:
            for a_uid in response.removedUsers:
                account_uid = utils.base64_url_encode(a_uid)
                if account_uid in params.user_cache:
                    del params.user_cache[account_uid]

        if len(response.users) > 0:
            for user in response.users:
                account_uid = utils.base64_url_encode(user.accountUid)
                params.user_cache[account_uid] = user.username
        account_uid = utils.base64_url_encode(params.account_uid_bytes)
        if account_uid not in params.user_cache:
            params.user_cache[account_uid] = params.user

        if len(response.recordRotations) > 0:
            for rr in response.recordRotations:
                record_uid = utils.base64_url_encode(rr.recordUid)
                rr_obj = {
                    'record_uid': record_uid,
                    'revision': rr.revision,
                    'configuration_uid': utils.base64_url_encode(rr.configurationUid),
                    'schedule': rr.schedule,
                    'pwd_complexity': utils.base64_url_encode(rr.pwdComplexity),
                    'disabled': rr.disabled,
                    'resource_uid': utils.base64_url_encode(rr.resourceUid),
                    'last_rotation': rr.lastRotation,
                    'last_rotation_status': rr.lastRotationStatus,
                }
                params.record_rotation_cache[record_uid] = rr_obj

        params.sync_down_token = response.continuationToken

    params.revision = revision

    for sf in params.shared_folder_cache.values():
        owner = sf.get('owner_username')
        if not owner:
            account_uid = sf.get('owner_account_uid')
            if account_uid and account_uid in params.user_cache:
                sf['owner_username'] = params.user_cache[account_uid]

        if 'users' in sf:
            for u in sf['users']:
                username = u.get('username')
                if not username:
                    account_uid = u['account_uid']
                    if account_uid and account_uid in params.user_cache:
                        u['username'] = params.user_cache[account_uid]

        if 'records' in sf:
            for r in sf['records']:
                owner = r.get('owner_username')
                if not owner:
                    account_uid = r['owner_account_uid']
                    if account_uid and account_uid in params.user_cache:
                        r['owner_username'] = params.user_cache[account_uid]

    for md in params.meta_data_cache.values():
        owner = md.get('owner_username')
        if not owner:
            account_uid = md.get('owner_account_uid')
            if account_uid and account_uid in params.user_cache:
                md['owner_username'] = params.user_cache[account_uid]

    if params.breach_watch_records:
        for bwr in params.breach_watch_records.values():
            scanned_by = bwr.get('scanned_by')
            if not scanned_by:
                account_uid = bwr.get('scanned_by_account_uid')
                if account_uid and account_uid in params.user_cache:
                    bwr['scanned_by'] = params.user_cache[account_uid]

    to_delete = set()

    logging.debug('Decrypting meta data keys')
    for record_uid, meta_data in params.meta_data_cache.items():
        if 'record_key_unencrypted' in meta_data:
            continue
        record_key = None
        try:
            if 'record_key' not in meta_data:
                # old record that doesn't have a record key so make one
                logging.debug('...no record key.  creating...')
                # store as b64 encoded string
                # note: decode() converts bytestream (b'') to string
                # note2: remove == from the end
                record_key = utils.generate_aes_key()
                record_key_encrypted = crypto.encrypt_aes_v1(record_key, params.data_key)
                meta_data['record_key'] = utils.base64_url_encode(record_key_encrypted)
                meta_data['record_key_type'] = 1
                # temporary flag for decryption routine below
                meta_data['old_record_flag'] = True
                meta_data['is_converted_record_type'] = True
            else:
                record_key_encrypted = utils.base64_url_decode(meta_data['record_key'])
                key_type = meta_data['record_key_type']
                if key_type == record_pb2.ENCRYPTED_BY_DATA_KEY:
                    record_key = crypto.decrypt_aes_v1(record_key_encrypted, params.data_key)
                elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY:
                    record_key = crypto.decrypt_rsa(record_key_encrypted, params.rsa_key2)
                elif key_type == record_pb2.ENCRYPTED_BY_DATA_KEY_GCM:
                    record_key = crypto.decrypt_aes_v2(record_key_encrypted, params.data_key)
                elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY_ECC:
                    record_key = crypto.decrypt_ec(record_key_encrypted, params.ecc_key)
                else:
                    raise Exception('Unsupported key type')
        except Exception as e:
            logging.debug('Record %s meta data decryption error: %s', record_uid, e)

        if record_key and len(record_key) == 32:
            meta_data['record_key_unencrypted'] = record_key
        else:
            to_delete.add(record_uid)

    for record_uid in to_delete:
        del params.meta_data_cache[record_uid]
    to_delete.clear()

    logging.debug('Decrypting team keys')
    for team_uid, team in params.team_cache.items():
        if 'team_key_unencrypted' not in team:
            try:
                encrypted_team_key = utils.base64_url_decode(team['team_key'])
                key_type = team['team_key_type']
                if key_type == record_pb2.ENCRYPTED_BY_DATA_KEY:
                    team_key = crypto.decrypt_aes_v1(encrypted_team_key, params.data_key)
                elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY:
                    team_key = crypto.decrypt_rsa(encrypted_team_key, params.rsa_key2)
                elif key_type == record_pb2.ENCRYPTED_BY_DATA_KEY_GCM:
                    team_key = crypto.decrypt_aes_v2(encrypted_team_key, params.data_key)
                elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY_ECC:
                    team_key = crypto.decrypt_ec(encrypted_team_key, params.ecc_key)
                else:
                    raise Exception('Unsupported key type')
                team['team_key_unencrypted'] = team_key
                if 'team_private_key' in team:
                    encrypted_team_private_key = utils.base64_url_decode(team['team_private_key'])
                    team['team_private_key_unencrypted'] = crypto.decrypt_aes_v1(encrypted_team_private_key, team_key)
                if 'team_ec_private_key' in team:
                    encrypted_team_private_key = utils.base64_url_decode(team['team_ec_private_key'])
                    team['team_ec_private_key_unencrypted'] = crypto.decrypt_aes_v2(encrypted_team_private_key, team_key)
            except Exception as e:
                logging.info('Could not decrypt team %s key: %s', team_uid, e)
        if 'team_key_unencrypted' in team:
            team_key = team['team_key_unencrypted']
            team_uid = team['team_uid']
            if 'shared_folder_keys' in team:
                for sf_key in team['shared_folder_keys']:
                    shared_folder_uid = sf_key['shared_folder_uid']
                    if 'shared_folder_key_unencrypted' not in sf_key:
                        encrypted_sf_key = utils.base64_url_decode(sf_key['shared_folder_key'])
                        try:
                            key_type = sf_key['key_type']
                            decrypted_sf_key = None
                            if key_type == record_pb2.ENCRYPTED_BY_DATA_KEY:
                                decrypted_sf_key = crypto.decrypt_aes_v1(encrypted_sf_key, team_key)
                            elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY:
                                if 'team_private_key_unencrypted' in team:
                                    team_private_key = team['team_private_key_unencrypted']
                                    team_pk = crypto.load_rsa_private_key(team_private_key)
                                    decrypted_sf_key = crypto.decrypt_rsa(encrypted_sf_key, team_pk)
                            elif key_type == record_pb2.ENCRYPTED_BY_DATA_KEY_GCM:
                                decrypted_sf_key = crypto.decrypt_aes_v2(encrypted_sf_key, team_key)
                            elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY_ECC:
                                if 'team_ec_private_key_unencrypted' in team:
                                    team_private_key = team['team_ec_private_key_unencrypted']
                                    team_pk = crypto.load_ec_private_key(team_private_key)
                                    decrypted_sf_key = crypto.decrypt_ec(encrypted_sf_key, team_pk)
                            else:
                                raise Exception('Unsupported key type')
                            if decrypted_sf_key:
                                sf_key['shared_folder_key_unencrypted'] = decrypted_sf_key
                            else:
                                logging.debug('Cannot decrypt team\' shared folder key: team_uid=%s, shared_folder_uid=%s', team_uid, shared_folder_uid)
                        except Exception as e:
                            logging.debug('Decryption error: team_uid=%s, shared_folder_uid=%s: %s', team_uid, shared_folder_uid, e)
        else:
            to_delete.add(team_uid)

    for team_uid in to_delete:
        del params.team_cache[team_uid]
    to_delete.clear()

    logging.debug('Decrypting shared folder keys')
    for shared_folder_uid, shared_folder in params.shared_folder_cache.items():
        if 'shared_folder_key_unencrypted' not in shared_folder and 'shared_folder_key' in shared_folder:
            # shared folder key
            try:
                encrypted_sf_key = utils.base64_url_decode(shared_folder['shared_folder_key'])
                key_type = shared_folder['key_type']
                if key_type == record_pb2.ENCRYPTED_BY_DATA_KEY:
                    sf_key = crypto.decrypt_aes_v1(encrypted_sf_key, params.data_key)
                elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY:
                    sf_key = crypto.decrypt_rsa(encrypted_sf_key, params.rsa_key2)
                elif key_type == record_pb2.ENCRYPTED_BY_DATA_KEY_GCM:
                    sf_key = crypto.decrypt_aes_v2(encrypted_sf_key, params.data_key)
                elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY_ECC:
                    sf_key = crypto.decrypt_ec(encrypted_sf_key, params.ecc_key)
                else:
                    sf_key = crypto.decrypt_aes_v1(encrypted_sf_key, params.data_key)
                shared_folder['shared_folder_key_unencrypted'] = sf_key
            except Exception as e:
                logging.debug('Shared folder %s key decryption error: %s', shared_folder_uid, e)

        if 'shared_folder_key_unencrypted' not in shared_folder:
            # team's shared folder key
            for team_uid, team in params.team_cache.items():
                if 'shared_folder_keys' in team:
                    sf_key = next((x for x in team['shared_folder_keys'] if x['shared_folder_uid'] == shared_folder_uid), None)
                    if sf_key and 'shared_folder_key_unencrypted' in sf_key:
                        shared_folder['shared_folder_key_unencrypted'] = sf_key['shared_folder_key_unencrypted']

        if 'shared_folder_key_unencrypted' in shared_folder:
            sf_key = shared_folder['shared_folder_key_unencrypted']
            try:
                if 'name_unencrypted' not in shared_folder:
                    name = shared_folder.get('name')
                    if name:
                        shared_folder['name_unencrypted'] = \
                            crypto.decrypt_aes_v1(utils.base64_url_decode(name), sf_key).decode('utf-8')
                    else:
                        data = shared_folder.get('data')
                        if data:
                            shared_folder['data_unencrypted'] = \
                                crypto.decrypt_aes_v1(utils.base64_url_decode(data), sf_key)
                            data_json = json.loads(shared_folder['data_unencrypted'].decode('utf-8'))
                            shared_folder['name_unencrypted'] = data_json['name']
                if 'data' in shared_folder and 'data_unencrypted' not in shared_folder:
                    data = utils.base64_url_decode(shared_folder['data'])
                    shared_folder['data_unencrypted'] = crypto.decrypt_aes_v1(data, sf_key)

            except Exception as e:
                logging.debug('Shared folder %s name decryption error: %s', shared_folder_uid, e)
            if 'name_unencrypted' not in shared_folder:
                shared_folder['name_unencrypted'] = shared_folder_uid

            if 'records' in shared_folder:
                for sfr in shared_folder['records']:
                    if 'record_key_unencrypted' not in sfr:
                        try:
                            encrypted_key = utils.base64_url_decode(sfr['record_key'])
                            if len(encrypted_key) == 60:
                                decrypted_key = crypto.decrypt_aes_v2(encrypted_key, sf_key)
                            else:
                                decrypted_key = crypto.decrypt_aes_v1(encrypted_key, sf_key)
                            sfr['record_key_unencrypted'] = decrypted_key
                        except Exception as e:
                            logging.debug('Shared folder %s record key decryption error: %s', shared_folder_uid, e)
        else:
            to_delete.add(shared_folder_uid)

    for shared_folder_uid in to_delete:
        del params.shared_folder_cache[shared_folder_uid]
        if shared_folder_uid in params.subfolder_cache:
            del params.subfolder_cache[shared_folder_uid]
    to_delete.clear()

    logging.debug('Resolve record keys. Meta data')
    for record_uid, record in params.record_cache.items():
        if 'record_key_unencrypted' not in record:
            # meta data
            if record_uid in params.meta_data_cache:
                meta_data = params.meta_data_cache[record_uid]
                if 'record_key_unencrypted' in meta_data:
                    record['record_key_unencrypted'] = meta_data['record_key_unencrypted']
        if 'record_key_unencrypted' not in record:
            to_delete.add(record_uid)

    if len(to_delete) > 0:
        logging.debug('Resolve record keys. Shared folder')
        for shared_folder in params.shared_folder_cache.values():
            if 'records' in shared_folder:
                for record_key in shared_folder['records']:
                    record_uid = record_key['record_uid']
                    if record_uid in to_delete and 'record_key_unencrypted' in record_key:
                        record = params.record_cache[record_uid]
                        record['record_key_unencrypted'] = record_key['record_key_unencrypted']
                        to_delete.remove(record_uid)
            if len(to_delete) == 0:
                break

    if len(to_delete) > 0:
        logging.debug('Decrypt linked keys')
        for child_uid, parents in params.record_link_cache.items():
            if child_uid in to_delete:
                for parent_uid, link in parents.items():
                    if 'record_key_unencrypted' not in link and parent_uid in params.record_cache:
                        parent_record = params.record_cache[parent_uid]
                        if 'record_key_unencrypted' in parent_record:
                            try:
                                encrypted_record_key = utils.base64_url_decode(link['record_key'])
                                link['record_key_unencrypted'] = crypto.decrypt_aes_v2(
                                    encrypted_record_key, parent_record['record_key_unencrypted'])
                            except Exception as e:
                                logging.debug('Record link %s key decryption error: %s', child_uid, e)

                child_record = params.record_cache[child_uid]
                if 'record_key_unencrypted' not in child_record:
                    for parent_uid, link in parents.items():
                        if 'record_key_unencrypted' in link:
                            child_record['record_key_unencrypted'] = link['record_key_unencrypted']
                            to_delete.remove(child_uid)
                            break

    to_delete.clear()
    for record_uid, record in params.record_cache.items():
        if 'record_key_unencrypted' not in record:
            logging.debug('Record %s key could not be resolved', record_uid)
            to_delete.add(record_uid)

    for record_uid in to_delete:
        if record_uid in params.record_link_cache:
            del params.record_link_cache[record_uid]
        for parents in params.record_link_cache.values():
            if record_uid in parents:
                del parents[record_uid]
        del params.record_cache[record_uid]
    to_delete.clear()

    logging.debug('Decrypting records')
    for record_uid, record in params.record_cache.items():
        record_key = record['record_key_unencrypted']
        if 'data_unencrypted' not in record:
            try:
                if 'version' in record and record['version'] >= 3:
                    record['data_unencrypted'] = crypto.decrypt_aes_v2(utils.base64_url_decode(record['data']), record_key) if 'data' in record else b'{}'
                else:
                    record['data_unencrypted'] = crypto.decrypt_aes_v1(utils.base64_url_decode(record['data']), record_key) if 'data' in record else b'{}'
                    extra = record.get('extra')
                    if extra:
                        record['extra_unencrypted'] = crypto.decrypt_aes_v1(utils.base64_url_decode(extra), record_key)
                    else:
                        record['extra_unencrypted'] = b'{}'
            except Exception as e:
                logging.debug('Record %s data/extra decryption error: %s', record_uid, e)

    logging.debug('Decrypting non shared data')
    for record_uid, nsd in params.non_shared_data_cache.items():
        if 'data_unencrypted' not in nsd and record_uid in params.record_cache:
            record = params.record_cache[record_uid]
            data = nsd.get('data')
            if data:
                version = record.get('version') or 0
                try:
                    if version >= 3:
                        nsd['data_unencrypted'] = crypto.decrypt_aes_v2(utils.base64_url_decode(data), params.data_key)
                    else:
                        nsd['data_unencrypted'] = crypto.decrypt_aes_v1(utils.base64_url_decode(data), params.data_key)
                except:
                    try:
                        if version < 3:
                            nsd['data_unencrypted'] = crypto.decrypt_aes_v2(utils.base64_url_decode(data), params.data_key)
                        else:
                            nsd['data_unencrypted'] = crypto.decrypt_aes_v1(utils.base64_url_decode(data), params.data_key)
                    except Exception as e:
                        logging.debug('Non Shared Data %s data decryption error: %s', record_uid, e)

    logging.debug('Decrypting folders')
    for folder_uid, sf in params.subfolder_cache.items():
        folder_type = sf['type']
        if folder_type == 'user_folder':
            if 'folder_key_unencrypted' not in sf:
                try:
                    encrypted_key = utils.base64_url_decode(sf['user_folder_key'])
                    key_type = sf['key_type']
                    if key_type == record_pb2.ENCRYPTED_BY_DATA_KEY:
                        sf['folder_key_unencrypted'] = crypto.decrypt_aes_v1(encrypted_key, params.data_key)
                    elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY:
                        sf['folder_key_unencrypted'] = crypto.decrypt_rsa(encrypted_key, params.rsa_key2)
                    elif key_type == record_pb2.ENCRYPTED_BY_DATA_KEY_GCM:
                        sf['folder_key_unencrypted'] = crypto.decrypt_aes_v2(encrypted_key, params.data_key)
                    elif key_type == record_pb2.ENCRYPTED_BY_PUBLIC_KEY_ECC:
                        sf['folder_key_unencrypted'] = crypto.decrypt_ec(encrypted_key, params.ecc_key)
                    else:
                        sf['folder_key_unencrypted'] = crypto.decrypt_aes_v1(encrypted_key, params.data_key)
                except Exception as e:
                    logging.debug('User folder data decryption error: %s', e)
        elif folder_type == 'shared_folder_folder':
            if 'folder_key_unencrypted' not in sf:
                try:
                    shared_folder_uid = sf['shared_folder_uid']
                    if shared_folder_uid in params.shared_folder_cache:
                        shared_folder = params.shared_folder_cache[shared_folder_uid]
                        encrypted_key = utils.base64_url_decode(sf['shared_folder_folder_key'])
                        sf['folder_key_unencrypted'] = crypto.decrypt_aes_v1(encrypted_key, shared_folder['shared_folder_key_unencrypted'])
                except Exception as e:
                    logging.debug('Shared folder folder %s data decryption error: %s', sf['folder_uid'], e)
        else:
            continue
        if 'folder_key_unencrypted' in sf:
            if 'data_unencrypted' not in sf:
                try:
                    data_encrypted = utils.base64_url_decode(sf['data'])
                    sf['data_unencrypted'] = crypto.decrypt_aes_v1(data_encrypted, sf['folder_key_unencrypted'])
                except Exception as e:
                    logging.debug('Error decrypting shared folder folder %s data: %s', sf['folder_uid'], e)

    prepare_folder_tree(params)

    # Populate/update cache record security data
    for sec_data in resp_sec_data_recs:
        record_uid = utils.base64_url_encode(sec_data.recordUid)
        params.breach_watch_security_data[record_uid] = {'revision': sec_data.revision}

    # Populate/update security score data
    for sec_score_rec in resp_sec_scores:
        record_uid = utils.base64_url_encode(sec_score_rec.recordUid or b'')
        record = KeeperRecord.load(params, record_uid)
        if not record:
            continue
        revision = sec_score_rec.revision
        data = sec_score_rec.data
        try:
            if data:
                data = crypto.decrypt_aes_v2(sec_score_rec.data, record.record_key).decode()
                data = json.loads(data)
            else:
                data = dict()
        except:
            data = dict()
        params.security_score_data.update({record_uid: dict(record_uid=record_uid, data=data, revision=revision)})

    # Populate/update cache BreachWatch records data
    for p_bwr in resp_bw_recs:
        record_uid = utils.base64_url_encode(p_bwr.recordUid)
        if not record_uid:
            continue
        record = params.record_cache.get(record_uid)
        if not record:
            continue
        if 'record_key_unencrypted' not in record:
            continue
        try:
            bwr = {
                'record_uid': utils.base64_url_encode(p_bwr.recordUid),
                'data': utils.base64_url_encode(p_bwr.data),
                'type': 'RECORD' if p_bwr.type == breachwatch_pb2.RECORD else 'ALTERNATE_PASSWORD',
                'revision': p_bwr.revision,
                'scanned_by_account_uid': utils.base64_url_encode(p_bwr.scannedByAccountUid or params.account_uid_bytes)
            }
            if len(p_bwr.data) > 0:
                data = crypto.decrypt_aes_v2(p_bwr.data, record['record_key_unencrypted'])
                data_obj = client_pb2.BreachWatchData()
                data_obj.ParseFromString(data)
                bwr['data_unencrypted'] = google.protobuf.json_format.MessageToDict(data_obj)
            params.breach_watch_records[record_uid] = bwr
        except Exception as e:
            logging.debug('Decrypt bw data: %s', e)

    if full_sync or record_types:
        # Record V3 types cache population
        record_types_rs = _sync_record_types(params)
        if len(record_types_rs.recordTypes) > 0:
            params.record_type_cache = {}
            for rt in record_types_rs.recordTypes:
                type_id = rt.recordTypeId
                type_id += rt.scope * 1000000
                params.record_type_cache[type_id] = rt.content

    if full_sync:
        convert_keys.change_key_types(params)

        if params.breach_watch:
            weak_count = 0
            for _ in params.breach_watch.get_records_by_status(params, ['WEAK', 'BREACHED']):
                weak_count += 1
            if weak_count > 0:
                logging.info(bcolors.WARNING +
                             f'The number of records that are affected by breaches or contain high-risk passwords: {weak_count}' +
                             '\nUse \"breachwatch list\" command to get more details' +
                             bcolors.ENDC)

        record_count = 0
        valid_versions = {2, 3}
        for r in params.record_cache.values():
            if r.get('version', 0) in valid_versions:
                record_count += 1
        if record_count:
            logging.info('Decrypted [%d] record(s)', record_count)


def _sync_record_types(params):  # type: (KeeperParams) -> Any
    rq = record_pb2.RecordTypesRequest()
    rq.standard = True
    rq.user = True
    rq.enterprise = True
    rq.pam = True
    return api.communicate_rest(params, rq, 'vault/get_record_types', rs_type=record_pb2.RecordTypesResponse)


def merge_lists_on_value(list1, list2, field_name):
    d = {x[field_name]: x for x in list1}
    d.update({x[field_name]: x for x in list2})
    return [x for x in d.values()]


def prepare_folder_tree(params):    # type: (KeeperParams) -> None
    params.folder_cache = {}
    params.root_folder = RootFolderNode()

    for sf in params.subfolder_cache.values():
        data_unencrypted = sf.get('data_unencrypted')    # type: Optional[bytes]
        folder_uid = None
        if sf['type'] == 'user_folder':
            folder_uid = sf['folder_uid']
            uf = UserFolderNode()
            uf.uid = folder_uid
            uf.parent_uid = sf.get('parent_uid')
            params.folder_cache[uf.uid] = uf

        elif sf['type'] == 'shared_folder_folder':
            folder_uid = sf['folder_uid']
            sff = SharedFolderFolderNode()
            sff.uid = folder_uid
            sff.shared_folder_uid = sf['shared_folder_uid']
            sff.parent_uid = sf.get('parent_uid') or sff.shared_folder_uid
            params.folder_cache[sff.uid] = sff

        elif sf['type'] == 'shared_folder':
            folder_uid = sf['shared_folder_uid']
            shf = SharedFolderNode()
            shf.uid = folder_uid
            shf.parent_uid = sf.get('folder_uid')
            folder = params.shared_folder_cache.get(shf.uid)
            if folder is not None:
                data_unencrypted = folder.get('data_unencrypted')
                shf.name = folder['name_unencrypted']
            params.folder_cache[shf.uid] = shf

        if data_unencrypted and folder_uid:
            try:
                f = params.folder_cache.get(folder_uid)    # type: Optional[BaseFolderNode]
                data = json.loads(data_unencrypted.decode())
                f.name = data.get('name') or f.name or f.uid
                f.color = data.get('color')
            except Exception as e:
                logging.debug('Error decrypting user folder name. Folder UID: %s. Error: %s', sf.uid, e)

    for f in params.folder_cache.values():
        parent_folder = params.folder_cache.get(f.parent_uid) if f.parent_uid else params.root_folder
        if parent_folder:
            parent_folder.subfolders.append(f.uid)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2019 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import logging

warned_on_fido_package = False
install_fido_package_warning = """
    You can use Security Key with Commander:
    Upgrade your Python interpreter to 3.10 or newer
    and make sure fido2 package is 2.0.0 or newer
"""

def display_fido2_warning():
    global warned_on_fido_package

    if not warned_on_fido_package:
        logging.warning(install_fido_package_warning)
    warned_on_fido_package = True



import APIRequest_pb2 as _APIRequest_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AuthProtocolType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    SAML2: _ClassVar[AuthProtocolType]

class DataType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    ANY: _ClassVar[DataType]
    BOOLEAN: _ClassVar[DataType]
    INTEGER: _ClassVar[DataType]
    STRING: _ClassVar[DataType]
    BYTES: _ClassVar[DataType]
    URL: _ClassVar[DataType]
    com_keepersecurity_proto_SsoCloud_DataType: _ClassVar[DataType]
    com_keepersecurity_proto_SsoCloud_AuthProtocolType: _ClassVar[DataType]
    com_keepersecurity_proto_SsoCloud_SsoIdpType: _ClassVar[DataType]
    LONG: _ClassVar[DataType]
    TIMESTAMP: _ClassVar[DataType]

class SsoCloudSettingOperationType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    SET: _ClassVar[SsoCloudSettingOperationType]
    GET: _ClassVar[SsoCloudSettingOperationType]
    DELETE: _ClassVar[SsoCloudSettingOperationType]
    RESET_TO_DEFAULT: _ClassVar[SsoCloudSettingOperationType]

class SsoIdpType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    XX_UNUSED: _ClassVar[SsoIdpType]
    GENERIC: _ClassVar[SsoIdpType]
    F5: _ClassVar[SsoIdpType]
    GOOGLE: _ClassVar[SsoIdpType]
    OKTA: _ClassVar[SsoIdpType]
    ADFS: _ClassVar[SsoIdpType]
    AZURE: _ClassVar[SsoIdpType]
    ONELOGIN: _ClassVar[SsoIdpType]
    AWS: _ClassVar[SsoIdpType]
    CENTRIFY: _ClassVar[SsoIdpType]
    DUO: _ClassVar[SsoIdpType]
    IBM: _ClassVar[SsoIdpType]
    JUMPCLOUD: _ClassVar[SsoIdpType]
    PING: _ClassVar[SsoIdpType]
    PINGONE: _ClassVar[SsoIdpType]
    RSA: _ClassVar[SsoIdpType]
    SECUREAUTH: _ClassVar[SsoIdpType]
    THALES: _ClassVar[SsoIdpType]
    AUTH0: _ClassVar[SsoIdpType]
    BEYOND: _ClassVar[SsoIdpType]
    HYPR: _ClassVar[SsoIdpType]
    PUREID: _ClassVar[SsoIdpType]
    SDO: _ClassVar[SsoIdpType]
    TRAIT: _ClassVar[SsoIdpType]
    TRANSMIT: _ClassVar[SsoIdpType]
    TRUSONA: _ClassVar[SsoIdpType]
    VERIDIUM: _ClassVar[SsoIdpType]
    CAS: _ClassVar[SsoIdpType]
SAML2: AuthProtocolType
ANY: DataType
BOOLEAN: DataType
INTEGER: DataType
STRING: DataType
BYTES: DataType
URL: DataType
com_keepersecurity_proto_SsoCloud_DataType: DataType
com_keepersecurity_proto_SsoCloud_AuthProtocolType: DataType
com_keepersecurity_proto_SsoCloud_SsoIdpType: DataType
LONG: DataType
TIMESTAMP: DataType
SET: SsoCloudSettingOperationType
GET: SsoCloudSettingOperationType
DELETE: SsoCloudSettingOperationType
RESET_TO_DEFAULT: SsoCloudSettingOperationType
XX_UNUSED: SsoIdpType
GENERIC: SsoIdpType
F5: SsoIdpType
GOOGLE: SsoIdpType
OKTA: SsoIdpType
ADFS: SsoIdpType
AZURE: SsoIdpType
ONELOGIN: SsoIdpType
AWS: SsoIdpType
CENTRIFY: SsoIdpType
DUO: SsoIdpType
IBM: SsoIdpType
JUMPCLOUD: SsoIdpType
PING: SsoIdpType
PINGONE: SsoIdpType
RSA: SsoIdpType
SECUREAUTH: SsoIdpType
THALES: SsoIdpType
AUTH0: SsoIdpType
BEYOND: SsoIdpType
HYPR: SsoIdpType
PUREID: SsoIdpType
SDO: SsoIdpType
TRAIT: SsoIdpType
TRANSMIT: SsoIdpType
TRUSONA: SsoIdpType
VERIDIUM: SsoIdpType
CAS: SsoIdpType

class SsoCloudSettingValue(_message.Message):
    __slots__ = ["settingId", "settingName", "label", "value", "valueType", "lastModified", "isFromFile", "isEditable", "isRequired"]
    SETTINGID_FIELD_NUMBER: _ClassVar[int]
    SETTINGNAME_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    VALUETYPE_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    ISFROMFILE_FIELD_NUMBER: _ClassVar[int]
    ISEDITABLE_FIELD_NUMBER: _ClassVar[int]
    ISREQUIRED_FIELD_NUMBER: _ClassVar[int]
    settingId: int
    settingName: str
    label: str
    value: str
    valueType: DataType
    lastModified: str
    isFromFile: bool
    isEditable: bool
    isRequired: bool
    def __init__(self, settingId: _Optional[int] = ..., settingName: _Optional[str] = ..., label: _Optional[str] = ..., value: _Optional[str] = ..., valueType: _Optional[_Union[DataType, str]] = ..., lastModified: _Optional[str] = ..., isFromFile: bool = ..., isEditable: bool = ..., isRequired: bool = ...) -> None: ...

class SsoCloudSettingAction(_message.Message):
    __slots__ = ["settingId", "settingName", "operation", "value"]
    SETTINGID_FIELD_NUMBER: _ClassVar[int]
    SETTINGNAME_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    settingId: int
    settingName: str
    operation: SsoCloudSettingOperationType
    value: str
    def __init__(self, settingId: _Optional[int] = ..., settingName: _Optional[str] = ..., operation: _Optional[_Union[SsoCloudSettingOperationType, str]] = ..., value: _Optional[str] = ...) -> None: ...

class SsoCloudConfigurationRequest(_message.Message):
    __slots__ = ["ssoServiceProviderId", "ssoSpConfigurationId", "name", "ssoAuthProtocolType", "ssoCloudSettingAction"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SSOAUTHPROTOCOLTYPE_FIELD_NUMBER: _ClassVar[int]
    SSOCLOUDSETTINGACTION_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    ssoSpConfigurationId: int
    name: str
    ssoAuthProtocolType: AuthProtocolType
    ssoCloudSettingAction: _containers.RepeatedCompositeFieldContainer[SsoCloudSettingAction]
    def __init__(self, ssoServiceProviderId: _Optional[int] = ..., ssoSpConfigurationId: _Optional[int] = ..., name: _Optional[str] = ..., ssoAuthProtocolType: _Optional[_Union[AuthProtocolType, str]] = ..., ssoCloudSettingAction: _Optional[_Iterable[_Union[SsoCloudSettingAction, _Mapping]]] = ...) -> None: ...

class SsoSharedConfigItem(_message.Message):
    __slots__ = ["ssoSpConfigurationId", "ssoServiceProviderId", "ssoNodeId"]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    SSONODEID_FIELD_NUMBER: _ClassVar[int]
    ssoSpConfigurationId: int
    ssoServiceProviderId: int
    ssoNodeId: int
    def __init__(self, ssoSpConfigurationId: _Optional[int] = ..., ssoServiceProviderId: _Optional[int] = ..., ssoNodeId: _Optional[int] = ...) -> None: ...

class SsoCloudConfigurationResponse(_message.Message):
    __slots__ = ["ssoServiceProviderId", "ssoSpConfigurationId", "enterpriseId", "name", "protocol", "lastModified", "ssoCloudSettingValue", "isShared", "sharedConfigs"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PROTOCOL_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    SSOCLOUDSETTINGVALUE_FIELD_NUMBER: _ClassVar[int]
    ISSHARED_FIELD_NUMBER: _ClassVar[int]
    SHAREDCONFIGS_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    ssoSpConfigurationId: int
    enterpriseId: int
    name: str
    protocol: str
    lastModified: str
    ssoCloudSettingValue: _containers.RepeatedCompositeFieldContainer[SsoCloudSettingValue]
    isShared: bool
    sharedConfigs: _containers.RepeatedCompositeFieldContainer[SsoSharedConfigItem]
    def __init__(self, ssoServiceProviderId: _Optional[int] = ..., ssoSpConfigurationId: _Optional[int] = ..., enterpriseId: _Optional[int] = ..., name: _Optional[str] = ..., protocol: _Optional[str] = ..., lastModified: _Optional[str] = ..., ssoCloudSettingValue: _Optional[_Iterable[_Union[SsoCloudSettingValue, _Mapping]]] = ..., isShared: bool = ..., sharedConfigs: _Optional[_Iterable[_Union[SsoSharedConfigItem, _Mapping]]] = ...) -> None: ...

class SsoIdpTypeRequest(_message.Message):
    __slots__ = ["ssoIdpTypeId", "tag", "label"]
    SSOIDPTYPEID_FIELD_NUMBER: _ClassVar[int]
    TAG_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    ssoIdpTypeId: int
    tag: str
    label: str
    def __init__(self, ssoIdpTypeId: _Optional[int] = ..., tag: _Optional[str] = ..., label: _Optional[str] = ...) -> None: ...

class SsoIdpTypeResponse(_message.Message):
    __slots__ = ["ssoIdpTypeId", "tag", "label"]
    SSOIDPTYPEID_FIELD_NUMBER: _ClassVar[int]
    TAG_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    ssoIdpTypeId: int
    tag: int
    label: int
    def __init__(self, ssoIdpTypeId: _Optional[int] = ..., tag: _Optional[int] = ..., label: _Optional[int] = ...) -> None: ...

class SsoCloudSAMLLogRequest(_message.Message):
    __slots__ = ["ssoServiceProviderId"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    def __init__(self, ssoServiceProviderId: _Optional[int] = ...) -> None: ...

class SsoCloudSAMLLogEntry(_message.Message):
    __slots__ = ["serverTime", "direction", "messageType", "messageIssued", "fromEntityId", "samlStatus", "relayState", "samlContent", "isSigned", "isOK"]
    SERVERTIME_FIELD_NUMBER: _ClassVar[int]
    DIRECTION_FIELD_NUMBER: _ClassVar[int]
    MESSAGETYPE_FIELD_NUMBER: _ClassVar[int]
    MESSAGEISSUED_FIELD_NUMBER: _ClassVar[int]
    FROMENTITYID_FIELD_NUMBER: _ClassVar[int]
    SAMLSTATUS_FIELD_NUMBER: _ClassVar[int]
    RELAYSTATE_FIELD_NUMBER: _ClassVar[int]
    SAMLCONTENT_FIELD_NUMBER: _ClassVar[int]
    ISSIGNED_FIELD_NUMBER: _ClassVar[int]
    ISOK_FIELD_NUMBER: _ClassVar[int]
    serverTime: str
    direction: str
    messageType: str
    messageIssued: str
    fromEntityId: str
    samlStatus: str
    relayState: str
    samlContent: str
    isSigned: bool
    isOK: bool
    def __init__(self, serverTime: _Optional[str] = ..., direction: _Optional[str] = ..., messageType: _Optional[str] = ..., messageIssued: _Optional[str] = ..., fromEntityId: _Optional[str] = ..., samlStatus: _Optional[str] = ..., relayState: _Optional[str] = ..., samlContent: _Optional[str] = ..., isSigned: bool = ..., isOK: bool = ...) -> None: ...

class SsoCloudSAMLLogResponse(_message.Message):
    __slots__ = ["ssoServiceProviderId", "entry"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    ENTRY_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    entry: _containers.RepeatedCompositeFieldContainer[SsoCloudSAMLLogEntry]
    def __init__(self, ssoServiceProviderId: _Optional[int] = ..., entry: _Optional[_Iterable[_Union[SsoCloudSAMLLogEntry, _Mapping]]] = ...) -> None: ...

class SsoCloudServiceProviderUpdateRequest(_message.Message):
    __slots__ = ["ssoServiceProviderId", "ssoSpConfigurationId"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    ssoSpConfigurationId: int
    def __init__(self, ssoServiceProviderId: _Optional[int] = ..., ssoSpConfigurationId: _Optional[int] = ...) -> None: ...

class SsoCloudIdpMetadataRequest(_message.Message):
    __slots__ = ["ssoSpConfigurationId", "filename", "content"]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    ssoSpConfigurationId: int
    filename: str
    content: bytes
    def __init__(self, ssoSpConfigurationId: _Optional[int] = ..., filename: _Optional[str] = ..., content: _Optional[bytes] = ...) -> None: ...

class SsoCloudIdpMetadataSupportRequest(_message.Message):
    __slots__ = ["ssoServiceProviderId", "ssoSpConfigurationId", "ssoEnterpriseId", "filename", "content"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    SSOENTERPRISEID_FIELD_NUMBER: _ClassVar[int]
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    ssoSpConfigurationId: int
    ssoEnterpriseId: int
    filename: str
    content: bytes
    def __init__(self, ssoServiceProviderId: _Optional[int] = ..., ssoSpConfigurationId: _Optional[int] = ..., ssoEnterpriseId: _Optional[int] = ..., filename: _Optional[str] = ..., content: _Optional[bytes] = ...) -> None: ...

class SsoCloudConfigurationValidationRequest(_message.Message):
    __slots__ = ["ssoSpConfigurationId"]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    ssoSpConfigurationId: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, ssoSpConfigurationId: _Optional[_Iterable[int]] = ...) -> None: ...

class ValidationContent(_message.Message):
    __slots__ = ["ssoSpConfigurationId", "isSuccessful", "errorMessage"]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    ISSUCCESSFUL_FIELD_NUMBER: _ClassVar[int]
    ERRORMESSAGE_FIELD_NUMBER: _ClassVar[int]
    ssoSpConfigurationId: int
    isSuccessful: bool
    errorMessage: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, ssoSpConfigurationId: _Optional[int] = ..., isSuccessful: bool = ..., errorMessage: _Optional[_Iterable[str]] = ...) -> None: ...

class SsoCloudConfigurationValidationResponse(_message.Message):
    __slots__ = ["validationContent"]
    VALIDATIONCONTENT_FIELD_NUMBER: _ClassVar[int]
    validationContent: _containers.RepeatedCompositeFieldContainer[ValidationContent]
    def __init__(self, validationContent: _Optional[_Iterable[_Union[ValidationContent, _Mapping]]] = ...) -> None: ...

class SsoCloudServiceProviderConfigurationListRequest(_message.Message):
    __slots__ = ["ssoServiceProviderId"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    def __init__(self, ssoServiceProviderId: _Optional[int] = ...) -> None: ...

class ConfigurationListItem(_message.Message):
    __slots__ = ["ssoSpConfigurationId", "name", "isSelected", "ssoServiceProviderId"]
    SSOSPCONFIGURATIONID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ISSELECTED_FIELD_NUMBER: _ClassVar[int]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    ssoSpConfigurationId: int
    name: str
    isSelected: bool
    ssoServiceProviderId: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, ssoSpConfigurationId: _Optional[int] = ..., name: _Optional[str] = ..., isSelected: bool = ..., ssoServiceProviderId: _Optional[_Iterable[int]] = ...) -> None: ...

class SsoCloudServiceProviderConfigurationListResponse(_message.Message):
    __slots__ = ["configurationItem"]
    CONFIGURATIONITEM_FIELD_NUMBER: _ClassVar[int]
    configurationItem: _containers.RepeatedCompositeFieldContainer[ConfigurationListItem]
    def __init__(self, configurationItem: _Optional[_Iterable[_Union[ConfigurationListItem, _Mapping]]] = ...) -> None: ...

class SsoCloudRequest(_message.Message):
    __slots__ = ["messageSessionUid", "clientVersion", "embedded", "json", "dest", "idpSessionId", "forceLogin", "username", "detached"]
    MESSAGESESSIONUID_FIELD_NUMBER: _ClassVar[int]
    CLIENTVERSION_FIELD_NUMBER: _ClassVar[int]
    EMBEDDED_FIELD_NUMBER: _ClassVar[int]
    JSON_FIELD_NUMBER: _ClassVar[int]
    DEST_FIELD_NUMBER: _ClassVar[int]
    IDPSESSIONID_FIELD_NUMBER: _ClassVar[int]
    FORCELOGIN_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    DETACHED_FIELD_NUMBER: _ClassVar[int]
    messageSessionUid: bytes
    clientVersion: str
    embedded: bool
    json: bool
    dest: str
    idpSessionId: str
    forceLogin: bool
    username: str
    detached: bool
    def __init__(self, messageSessionUid: _Optional[bytes] = ..., clientVersion: _Optional[str] = ..., embedded: bool = ..., json: bool = ..., dest: _Optional[str] = ..., idpSessionId: _Optional[str] = ..., forceLogin: bool = ..., username: _Optional[str] = ..., detached: bool = ...) -> None: ...

class SsoCloudResponse(_message.Message):
    __slots__ = ["command", "messageSessionUid", "email", "encryptedLoginToken", "providerName", "idpSessionId", "encryptedSessionToken", "errorToken"]
    COMMAND_FIELD_NUMBER: _ClassVar[int]
    MESSAGESESSIONUID_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDLOGINTOKEN_FIELD_NUMBER: _ClassVar[int]
    PROVIDERNAME_FIELD_NUMBER: _ClassVar[int]
    IDPSESSIONID_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDSESSIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    ERRORTOKEN_FIELD_NUMBER: _ClassVar[int]
    command: str
    messageSessionUid: bytes
    email: str
    encryptedLoginToken: bytes
    providerName: str
    idpSessionId: str
    encryptedSessionToken: bytes
    errorToken: str
    def __init__(self, command: _Optional[str] = ..., messageSessionUid: _Optional[bytes] = ..., email: _Optional[str] = ..., encryptedLoginToken: _Optional[bytes] = ..., providerName: _Optional[str] = ..., idpSessionId: _Optional[str] = ..., encryptedSessionToken: _Optional[bytes] = ..., errorToken: _Optional[str] = ...) -> None: ...

class SsoCloudLogRequest(_message.Message):
    __slots__ = ["ssoServiceProviderId", "serviceName", "serviceId"]
    SSOSERVICEPROVIDERID_FIELD_NUMBER: _ClassVar[int]
    SERVICENAME_FIELD_NUMBER: _ClassVar[int]
    SERVICEID_FIELD_NUMBER: _ClassVar[int]
    ssoServiceProviderId: int
    serviceName: str
    serviceId: int
    def __init__(self, ssoServiceProviderId: _Optional[int] = ..., serviceName: _Optional[str] = ..., serviceId: _Optional[int] = ...) -> None: ...

class SamlRelayState(_message.Message):
    __slots__ = ["messageSessionUid", "username", "embedded", "json", "destId", "keyId", "supportedLanguage", "checksum", "isGeneratedUid", "deviceId", "detached"]
    MESSAGESESSIONUID_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    EMBEDDED_FIELD_NUMBER: _ClassVar[int]
    JSON_FIELD_NUMBER: _ClassVar[int]
    DESTID_FIELD_NUMBER: _ClassVar[int]
    KEYID_FIELD_NUMBER: _ClassVar[int]
    SUPPORTEDLANGUAGE_FIELD_NUMBER: _ClassVar[int]
    CHECKSUM_FIELD_NUMBER: _ClassVar[int]
    ISGENERATEDUID_FIELD_NUMBER: _ClassVar[int]
    DEVICEID_FIELD_NUMBER: _ClassVar[int]
    DETACHED_FIELD_NUMBER: _ClassVar[int]
    messageSessionUid: bytes
    username: str
    embedded: bool
    json: bool
    destId: int
    keyId: int
    supportedLanguage: _APIRequest_pb2.SupportedLanguage
    checksum: int
    isGeneratedUid: bool
    deviceId: int
    detached: bool
    def __init__(self, messageSessionUid: _Optional[bytes] = ..., username: _Optional[str] = ..., embedded: bool = ..., json: bool = ..., destId: _Optional[int] = ..., keyId: _Optional[int] = ..., supportedLanguage: _Optional[_Union[_APIRequest_pb2.SupportedLanguage, str]] = ..., checksum: _Optional[int] = ..., isGeneratedUid: bool = ..., deviceId: _Optional[int] = ..., detached: bool = ...) -> None: ...

class SsoCloudMigrationStatusRequest(_message.Message):
    __slots__ = ["nodeId", "fullStatus", "includeMigratedUsers", "limit"]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    FULLSTATUS_FIELD_NUMBER: _ClassVar[int]
    INCLUDEMIGRATEDUSERS_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    nodeId: int
    fullStatus: bool
    includeMigratedUsers: bool
    limit: int
    def __init__(self, nodeId: _Optional[int] = ..., fullStatus: bool = ..., includeMigratedUsers: bool = ..., limit: _Optional[int] = ...) -> None: ...

class SsoCloudMigrationStatusResponse(_message.Message):
    __slots__ = ["success", "message", "nodeId", "ssoConnectId", "ssoConnectName", "ssoConnectCloudId", "ssoConnectCloudName", "totalUsersCount", "usersMigratedCount", "migratedUsers", "unmigratedUsers"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    SSOCONNECTID_FIELD_NUMBER: _ClassVar[int]
    SSOCONNECTNAME_FIELD_NUMBER: _ClassVar[int]
    SSOCONNECTCLOUDID_FIELD_NUMBER: _ClassVar[int]
    SSOCONNECTCLOUDNAME_FIELD_NUMBER: _ClassVar[int]
    TOTALUSERSCOUNT_FIELD_NUMBER: _ClassVar[int]
    USERSMIGRATEDCOUNT_FIELD_NUMBER: _ClassVar[int]
    MIGRATEDUSERS_FIELD_NUMBER: _ClassVar[int]
    UNMIGRATEDUSERS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    nodeId: int
    ssoConnectId: int
    ssoConnectName: str
    ssoConnectCloudId: int
    ssoConnectCloudName: str
    totalUsersCount: int
    usersMigratedCount: int
    migratedUsers: _containers.RepeatedCompositeFieldContainer[SsoCloudMigrationUserInfo]
    unmigratedUsers: _containers.RepeatedCompositeFieldContainer[SsoCloudMigrationUserInfo]
    def __init__(self, success: bool = ..., message: _Optional[str] = ..., nodeId: _Optional[int] = ..., ssoConnectId: _Optional[int] = ..., ssoConnectName: _Optional[str] = ..., ssoConnectCloudId: _Optional[int] = ..., ssoConnectCloudName: _Optional[str] = ..., totalUsersCount: _Optional[int] = ..., usersMigratedCount: _Optional[int] = ..., migratedUsers: _Optional[_Iterable[_Union[SsoCloudMigrationUserInfo, _Mapping]]] = ..., unmigratedUsers: _Optional[_Iterable[_Union[SsoCloudMigrationUserInfo, _Mapping]]] = ...) -> None: ...

class SsoCloudMigrationUserInfo(_message.Message):
    __slots__ = ["userId", "email", "fullName", "isMigrated"]
    USERID_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    FULLNAME_FIELD_NUMBER: _ClassVar[int]
    ISMIGRATED_FIELD_NUMBER: _ClassVar[int]
    userId: int
    email: str
    fullName: str
    isMigrated: bool
    def __init__(self, userId: _Optional[int] = ..., email: _Optional[str] = ..., fullName: _Optional[str] = ..., isMigrated: bool = ...) -> None: ...
